use std::cmp::Ordering::Less;
use std::fmt::{Debug, Display};
use num::Integer;

#[derive(Debug, PartialEq, Eq, PartialOrd, Clone, Copy)]
pub struct Point<T: Integer> {
    pub x: T,
    pub y: T,
}

impl<T: Integer> Point<T> {
    pub fn is_zero(&self) -> bool {
        self.x.is_zero() && self.y.is_zero()
    }
    pub fn is_negative(&self) -> bool {
        if self.x.cmp(&T::zero()) == Less { return true }
        if self.y.cmp(&T::zero()) == Less { return true }
        false
    }
}

impl<T: Integer + std::ops::Neg<Output = T> + Copy> Point<T> {
    pub fn distance(&self, other: &Self) -> T {
        let mut a = self.x - other.x;
        let mut b = self.y - other.y;
        if a.cmp(&T::zero()) == Less {
            a = -a;
        }
        if b.cmp(&T::zero()) == Less {
            b = -b;
        }
        a + b
    }
}

impl<T: Integer + Default + Copy> Point<T> {
    pub fn abs(&mut self) {
        if self.x.cmp(&T::zero()) == Less {
            let mut t = T::default();
            t.inc();
            t.inc();
            self.x = self.x - (self.x * t);
        }
        if self.y.cmp(&T::zero()) == Less {
            let mut t = T::default();
            t.inc();
            t.inc();
            self.y = self.y - (self.y * t);
        }
    }
}

impl<T: Integer + Copy> From<T> for Point<T> {
    fn from(value: T) -> Self {
        Self { x: value, y: value }
    }
}

impl<T: Integer> From<(T, T)> for Point<T> {
    fn from(value: (T, T)) -> Self {
        Self { x: value.0, y: value.1 }
    }
}

impl<T: Integer + Copy> From<&Point<T>> for Point<T> {
    fn from(value: &Point<T>) -> Self {
        Self { x: value.x, y: value.y }
    }
}


impl<T: Integer + Copy> From<&mut Point<T>> for Point<T> {
    fn from(value: &mut Point<T>) -> Self {
        Self { x: value.x, y: value.y }
    }
}

impl<T: Integer + Default> Default for Point<T> {
    fn default() -> Self {
        Self { x: T::default(), y: T::default() }
    }
}

impl<T: Integer + Display> Display for Point<T> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_str(format!("({}, {})", self.x, self.y).as_str())
    }
}
